#!/usr/bin/env python3

import os
from gi.repository import Gtk, Gdk, GLib
from i3ipc import Event

from nwg_panel.tools import check_key, get_icon_name, update_image, load_autotiling, get_config_dir, temp_dir, \
    save_json, update_image, update_image_fallback_desktop
import nwg_panel.common


class SwayTaskbar(Gtk.Box):
    def __init__(self, settings, i3, position, display_name="", icons_path=""):
        self.position = position
        self.icons_path = icons_path
        check_key(settings, "workspaces-spacing", 0)
        check_key(settings, "image-size", 16)
        check_key(settings, "workspace-menu", [1, 2, 3, 4, 5, 6, 7, 8])
        check_key(settings, "task-padding", 0)
        check_key(settings, "all-workspaces", True)
        check_key(settings, "mark-autotiling", True)
        check_key(settings, "mark-xwayland", True)
        check_key(settings, "angle", 0.0)

        self.cache_file = os.path.join(temp_dir(), "nwg-scratchpad")

        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=settings["workspaces-spacing"])
        self.settings = settings
        if self.settings["angle"] != 0.0:
            self.set_orientation(Gtk.Orientation.VERTICAL)

        self.display_name = display_name
        self.i3 = i3
        self.tree = i3.get_tree()
        self.displays_tree = self.list_tree()

        self.autotiling = load_autotiling() if settings["mark-autotiling"] else []

        self.build_box()
        self.ws_box = None
        self.subscribe()

    def subscribe(self):
        self.i3.on(Event.WINDOW, self.on_i3ipc_event)
        self.i3.on(Event.WORKSPACE, self.on_i3ipc_event)

    def list_tree(self):
        """
        display -> workspace -> window -> app_id
                                       -> parent_layout
                                       -> name
                                       -> pid
                                       -> con
                             -> window -> (...)
                -> workspace -> (...)
        display -> (...)
        """
        displays_tree = []
        if self.display_name:
            for item in self.tree:
                if item.type == "output" and item.name == self.display_name:
                    displays_tree.append(item)
        else:
            for item in self.tree:
                if item.type == "output" and not item.name.startswith("__"):
                    displays_tree.append(item)

        # sort by x, y coordinates
        displays_tree = sorted(displays_tree, key=lambda d: (d.rect.x, d.rect.y))

        return displays_tree

    def build_box(self):
        self.displays_tree = self.list_tree()
        all_workspaces = self.settings["all-workspaces"]

        for display in self.displays_tree:
            for desc in display.descendants():
                if desc.type == "workspace":
                    self.ws_box = WorkspaceBox(desc, self.settings, self.autotiling)
                    if all_workspaces or desc.find_focused() is not None:
                        for con in desc.descendants():
                            if con.name or con.app_id:
                                win_box = WindowBox(self.tree, con, self.settings, self.position, self.icons_path,
                                                    self.cache_file, floating=con in desc.floating_nodes)
                                self.ws_box.pack_start(win_box, False, False, self.settings["task-padding"])
                    self.pack_start(self.ws_box, False, False, 0)
        self.show_all()

    def on_i3ipc_event(self, i3conn, event):
        GLib.idle_add(self.refresh,
                      i3conn.get_tree(),
                      priority=GLib.PRIORITY_HIGH)

    def refresh(self, tree):
        self.tree = tree
        for item in self.get_children():
            item.destroy()
        self.build_box()


class WorkspaceBox(Gtk.Box):
    def __init__(self, con, settings, autotiling):
        self.con = con
        at_indicator = "a" if con.num in autotiling else ""
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        if settings["angle"] != 0.0:
            self.set_orientation(Gtk.Orientation.VERTICAL)

        check_key(settings, "workspace-buttons", False)
        if settings["workspace-buttons"]:
            widget = Gtk.Button.new_with_label("{}{}".format(at_indicator, con.num))
            widget.connect("clicked", self.on_click)
        else:
            widget = Gtk.Label("{}{}:".format(at_indicator, con.num))
            widget.set_angle(settings["angle"])

        self.pack_start(widget, False, False, 4)

    def on_click(self, button):
        nwg_panel.common.i3.command("{} number {} focus".format(self.con.type, self.con.num))


class WindowBox(Gtk.EventBox):
    def __init__(self, tree, con, settings, position, icons_path, cache_file, floating=False):
        self.position = position
        self.settings = settings
        Gtk.EventBox.__init__(self)
        self.box = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, spacing=0)
        if settings["angle"] != 0.0:
            self.box.set_orientation(Gtk.Orientation.VERTICAL)
        self.add(self.box)
        self.con = con
        self.pid = con.pid
        self.icons_path = icons_path
        self.tree = tree
        self.cache_file = cache_file

        self.old_name = ""

        if con.urgent:
            self.box.set_property("name", "task-box-urgent")
        elif con.focused:
            self.box.set_property("name", "task-box-focused")
        else:
            self.box.set_property("name", "task-box")

        self.connect('enter-notify-event', self.on_enter_notify_event)
        self.connect('leave-notify-event', self.on_leave_notify_event)
        self.connect('button-release-event', self.on_click, self.box)
        self.add_events(Gdk.EventMask.SCROLL_MASK)
        self.connect('scroll-event', self.on_scroll)

        check_key(settings, "show-app-icon", True)
        if settings["show-app-icon"]:
            name = con.app_id if con.app_id else con.window_class

            image = Gtk.Image()
            image.set_property("name", "task-box-icon")
            update_image_fallback_desktop(image, name, settings["image-size"], icons_path)
            self.box.pack_start(image, False, False, 4)

        if con.name:
            check_key(settings, "show-app-name", True)
            check_key(settings, "name-max-len", 20)
            name = con.name[:settings["name-max-len"]] if len(con.name) > settings["name-max-len"] else con.name
            if settings["mark-xwayland"] and not con.app_id:
                name = "X|" + name
            if settings["show-app-name"]:
                check_key(settings, "name-max-len", 10)
                label = Gtk.Label(name)
                label.set_angle(settings["angle"])
                self.box.pack_start(label, False, False, 0)
            else:
                self.set_tooltip_text(name)

        check_key(settings, "show-layout", True)

        if settings["show-layout"] and con.parent.layout:
            if con.parent.layout == "splith":
                image = Gtk.Image()
                update_image(image, "go-next-symbolic", 16, icons_path)
            elif con.parent.layout == "splitv":
                image = Gtk.Image()
                update_image(image, "go-down-symbolic", 16, icons_path)
            elif con.parent.layout == "tabbed":
                image = Gtk.Image()
                update_image(image, "view-dual-symbolic", 16, icons_path)
            elif con.parent.layout == "stacked":
                image = Gtk.Image()
                update_image(image, "view-paged-symbolic", 16, icons_path)

            if floating:
                image = Gtk.Image()
                update_image(image, "window-pop-out-symbolic", 16, icons_path)

            self.box.pack_start(image, False, False, 4)

    def on_enter_notify_event(self, widget, event):
        widget.set_state_flags(Gtk.StateFlags.DROP_ACTIVE, clear=False)
        widget.set_state_flags(Gtk.StateFlags.SELECTED, clear=False)

    def on_leave_notify_event(self, widget, event):
        widget.unset_state_flags(Gtk.StateFlags.DROP_ACTIVE)
        widget.unset_state_flags(Gtk.StateFlags.SELECTED)

    def on_click(self, widget, event, at_widget):
        if event.button == 1:
            cmd = "[con_id=\"{}\"] focus".format(self.con.id)
            nwg_panel.common.i3.command(cmd)
        if event.button == 3:
            menu = self.context_menu(self.settings["workspace-menu"])
            menu.show_all()
            if self.position == "bottom":
                menu.popup_at_widget(at_widget, Gdk.Gravity.SOUTH, Gdk.Gravity.NORTH, None)
            else:
                menu.popup_at_widget(at_widget, Gdk.Gravity.NORTH, Gdk.Gravity.SOUTH, None)

    def on_scroll(self, widget, event):
        cmd = "[con_id=\"{}\"] focus".format(self.con.id)
        nwg_panel.common.i3.command(cmd)
        if event.direction == Gdk.ScrollDirection.UP:
            cmd = "[con_id=\"{}\"] layout toggle tabbed stacking splitv splith".format(self.con.id)
            nwg_panel.common.i3.command(cmd)
        elif event.direction == Gdk.ScrollDirection.DOWN:
            cmd = "[con_id=\"{}\"] layout toggle splith splitv stacking tabbed".format(self.con.id)
            nwg_panel.common.i3.command(cmd)

    def context_menu(self, workspaces):
        menu = Gtk.Menu()
        menu.set_reserve_toggle_size(False)
        # Numbers have been converted to strings by mistake (config.py). Reverting this would be a breaking change,
        # so let's accept both numbers and strings.
        for i in workspaces:
            ws_num = self.con_ws_num(self.con)
            if str(i) != str(ws_num):
                hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
                img = Gtk.Image()
                update_image(img, "go-next-symbolic", 16, self.icons_path)
                hbox.pack_start(img, True, True, 0)
                label = Gtk.Label(str(i))
                hbox.pack_start(label, True, True, 0)
                item = Gtk.MenuItem()
                item.add(hbox)
                item.connect("activate", self.move_to_workspace, i)
                item.set_tooltip_text("move to workspace number {}".format(i))
                menu.append(item)

        # Move to scratchpad
        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "go-next-symbolic", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        img = Gtk.Image()
        update_image(img, "edit-paste", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.move_scratchpad)
        item.set_tooltip_text('move scratchpad ("minimize")')
        menu.append(item)

        item = Gtk.SeparatorMenuItem()

        menu.append(item)

        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "view-paged-symbolic", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        img = Gtk.Image()
        update_image(img, "view-dual-symbolic", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.floating_toggle)
        item.set_tooltip_text("floating toggle")
        menu.append(item)

        hbox = Gtk.Box.new(Gtk.Orientation.HORIZONTAL, 0)
        img = Gtk.Image()
        update_image(img, "window-close-symbolic", 16, self.icons_path)
        hbox.pack_start(img, True, True, 0)
        item = Gtk.MenuItem()
        item.add(hbox)
        item.connect("activate", self.kill)
        item.set_tooltip_text("kill")
        menu.append(item)

        return menu

    def con_ws_num(self, con):
        ws_num = 0
        leave = con
        parent = con.parent
        while parent.type != "root":
            parent = leave.parent
            leave = parent
            if leave.type == "workspace":
                ws_num = leave.num

        return ws_num

    def move_to_workspace(self, item, ws_num):
        cmd = "[con_id=\"{}\"] move to workspace number {}".format(self.con.id, ws_num)
        nwg_panel.common.i3.command(cmd)

        cmd = "[con_id=\"{}\"] focus".format(self.con.id)
        nwg_panel.common.i3.command(cmd)

    def move_scratchpad(self, item):
        # d dictionary is to remember con workspace number, floating state and parent output name,
        # to bring it back from scratchpad to its original place, in original state.
        d = {"floating_con": self.con.type == "floating_con"}
        for ws in self.tree.workspaces():
            if ws.find_by_id(self.con.id):
                d["workspace"] = ws.num
                break

        # We'll need the output name to (optionally) filter the Scratchpad module icons by the panel output name.
        d["output"] = self.con_parent_output_name(self.con)

        # key must not be a number, as json keys are always strings, and we will try to restore this dict from cache
        nwg_panel.common.scratchpad_cons[str(self.con.id)] = d

        # save to the cache file, for the information to survive panel restarts
        save_json(nwg_panel.common.scratchpad_cons, self.cache_file)

        cmd = "[con_id=\"{}\"] move to scratchpad".format(self.con.id)
        nwg_panel.common.i3.command(cmd)

    def floating_toggle(self, item):
        cmd = "[con_id=\"{}\"] floating toggle".format(self.con.id)
        nwg_panel.common.i3.command(cmd)

    def kill(self, item):
        cmd = "[con_id=\"{}\"] kill".format(self.con.id)
        nwg_panel.common.i3.command(cmd)

    def con_parent_output_name(self, con):
        p = con.parent
        if p:
            if p.type == "output":
                return p.name
            else:
                return self.con_parent_output_name(p)

        return None



from gi.repository import Gdk

from dasbus.connection import SessionMessageBus
from dasbus.specification import DBusSpecificationParser
from dasbus.client.observer import DBusObserver
from dasbus.client.proxy import disconnect_proxy
from dasbus.error import DBusError

from nwg_panel.tools import load_resource

PROPERTIES = [
    "Id",
    "Category",
    "Title",
    "Status",
    "WindowId",
    "IconName",
    "IconPixmap",
    "OverlayIconName",
    "OverlayIconPixmap",
    "AttentionIconName",
    "AttentionIconPixmap",
    "AttentionMovieName",
    "ToolTip",
    "IconThemePath",
    "ItemIsMenu",
    "Menu"
]


class StatusNotifierItem(object):
    def __init__(self, service_name, object_path):
        self.service_name = service_name
        self.object_path = object_path
        self.on_loaded_callback = None
        self.on_updated_callback = None
        self.session_bus = SessionMessageBus()
        self.properties = {
            "ItemIsMenu": True
        }
        self.item_proxy = None

        self.item_observer = DBusObserver(
            message_bus=self.session_bus,
            service_name=self.service_name
        )
        self.item_observer.service_available.connect(
            self.item_available_handler
        )
        self.item_observer.service_unavailable.connect(
            self.item_unavailable_handler
        )
        self.item_observer.connect_once_available()

    def __del__(self):
        if self.item_proxy is not None:
            disconnect_proxy(self.item_proxy)
        self.item_observer.disconnect()
        self.session_bus.disconnect()

    def item_available_handler(self, _observer):
        self.item_proxy = self.session_bus.get_proxy(self.service_name, self.object_path)
        try:
            spec = self.item_proxy._handler.specification
            if spec is not None:
                if not any("StatusNotifierItem" in ifname for ifname in spec.interfaces):
                    DBusSpecificationParser._parse_xml(spec, load_resource(__package__,"org.kde.StatusNotifierItem.xml"))
        except:
            pass
        if hasattr(self.item_proxy, "PropertiesChanged"):
            self.item_proxy.PropertiesChanged.connect(
                lambda _if, changed, invalid: self.change_handler(list(changed), invalid)
            )
        if hasattr(self.item_proxy, 'NewTitle'):
            self.item_proxy.NewTitle.connect(
                lambda: self.change_handler(["Title"])
            )
        if hasattr(self.item_proxy, 'NewToolTip'):
            self.item_proxy.NewToolTip.connect(
                lambda: self.change_handler(["ToolTip"])
            )
        if hasattr(self.item_proxy, 'NewIcon'):
            self.item_proxy.NewIcon.connect(
                lambda: self.change_handler(["IconName", "IconPixmap", "IconThemePath"])
            )
        if hasattr(self.item_proxy, 'NewAttentionIcon'):
            self.item_proxy.NewAttentionIcon.connect(
                lambda: self.change_handler(["AttentionIconName", "AttentionIconPixmap"])
            )
        if hasattr(self.item_proxy, "NewIconThemePath"):
            self.item_proxy.NewIconThemePath.connect(
                lambda _icon_theme_path: self.change_handler(["IconThemePath"])
            )
        if hasattr(self.item_proxy, "NewStatus"):
            self.item_proxy.NewStatus.connect(
                lambda _status: self.change_handler(["Status"])
            )
        for name in PROPERTIES:
            try:
                self.properties[name] = getattr(self.item_proxy, name)
            except (AttributeError, DBusError):
                # remote StatusNotifierItem object does not support all SNI properties
                pass
        if self.on_loaded_callback is not None:
            self.on_loaded_callback(self)

    def item_unavailable_handler(self, _observer):
        disconnect_proxy(self.item_proxy)
        self.item_proxy = None

    def change_handler(self, changed_properties: list[str], invalid_properties: list[str] = None):
        if invalid_properties is None:
            invalid_properties = []
        actual_changed_properties = []
        if len(changed_properties) > 0:
            for name in changed_properties:
                try:
                    self.properties[name] = getattr(self.item_proxy, name)
                    actual_changed_properties.append(name)
                except (AttributeError, DBusError):
                    pass
        if len(invalid_properties) > 0:
            for name in invalid_properties:
                if name in self.properties:
                    self.properties.pop(name)
        if len(actual_changed_properties) > 0:
            if self.on_updated_callback is not None:
                self.on_updated_callback(self, actual_changed_properties)

    def set_on_loaded_callback(self, callback):
        self.on_loaded_callback = callback

    def set_on_updated_callback(self, callback):
        self.on_updated_callback = callback

    @property
    def item_is_menu(self):
        if "ItemIsMenu" in self.properties:
            return self.properties["ItemIsMenu"]
        else:
            return False

    def context_menu(self, event: Gdk.EventButton):
        self.item_proxy.ContextMenu(event.x, event.y)

    def activate(self, event: Gdk.EventButton):
        self.item_proxy.Activate(event.x, event.y)

    def secondary_action(self, event: Gdk.EventButton):
        self.item_proxy.SecondaryAction(event.x, event.y)

    def scroll(self, distance, direction):
        self.item_proxy.Scroll(distance, direction)



#!/usr/bin/python3

"""
GTK3-based panel for sway Wayland compositor
Project: https://github.com/nwg-piotr/nwg-panel
Author's email: nwg.piotr@gmail.com
Copyright (c) 2021-2023 Piotr Miller & Contributors
License: MIT
"""
import argparse
import os
import signal
import sys
import threading

import gi

from nwg_panel.__about__ import __version__
from nwg_panel.modules.pinned import Pinned

gi.require_version('Gtk', '3.0')
try:
    gi.require_version('GtkLayerShell', '0.1')
except ValueError:

    raise RuntimeError('\n\n' +
                       'If you haven\'t installed GTK Layer Shell, you need to point Python to the\n' +
                       'library by setting GI_TYPELIB_PATH and LD_LIBRARY_PATH to <build-dir>/src/.\n' +
                       'For example you might need to run:\n\n' +
                       'GI_TYPELIB_PATH=build/src LD_LIBRARY_PATH=build/src python3 ' + ' '.join(sys.argv))

from gi.repository import GtkLayerShell, GLib

try:
    import psutil
except ModuleNotFoundError:
    print("You need to install python-psutil package")
    sys.exit(1)

from nwg_panel.tools import *

from nwg_panel.modules.custom_button import CustomButton
from nwg_panel.modules.executor import Executor
from nwg_panel.modules.clock import Clock
from nwg_panel.modules.controls import Controls
from nwg_panel.modules.playerctl import Playerctl
from nwg_panel.modules.cpu_avg import CpuAvg
from nwg_panel.modules.scratchpad import Scratchpad
from nwg_panel.modules.dwl_tags import DwlTags
from nwg_panel.modules.swaync import SwayNC
from nwg_panel.modules.sway_mode import SwayMode
from nwg_panel.modules.hyprland_submap import HyprlandSubmap
from nwg_panel.modules.keyboard_layout import KeyboardLayout
from nwg_panel.modules.random_wallpaper import RandomWallpaper

try:
    from nwg_panel.modules.openweather import OpenWeather
except Exception as e:
    eprint("Couldn't load OpenWeather module: ".format(e))
from nwg_panel.modules.brightness_slider import BrightnessSlider

from nwg_panel.modules.menu_start import MenuStart

dir_name = os.path.dirname(__file__)

from nwg_panel import common

tray_available = False
try:
    from nwg_panel.modules import sni_system_tray

    tray_available = True
except:
    eprint("Couldn't load system tray, is 'python-dasbus' installed?")

sway = os.getenv('SWAYSOCK') is not None
if sway:
    try:
        import i3ipc
        from i3ipc import Connection, Event
    except ModuleNotFoundError:
        eprint("'python-i3ipc' package required on sway, terminating")
        sys.exit(1)

    common.i3 = Connection()
    from nwg_panel.modules.sway_taskbar import SwayTaskbar
    from nwg_panel.modules.sway_workspaces import SwayWorkspaces

his = os.getenv('HYPRLAND_INSTANCE_SIGNATURE')
if his:
    from nwg_panel.modules.hyprland_taskbar import HyprlandTaskbar
    from nwg_panel.modules.hyprland_workspaces import HyprlandWorkspaces

niri_sock = os.getenv('NIRI_SOCKET')
if niri_sock:
    from nwg_panel.modules.niri_taskbar import NiriTaskbar

common_settings = {}
restart_cmd = ""
sig_dwl = 0
voc = {}

panel_windows_hide_show_sigs = {}


def load_vocabulary():
    global voc
    # basic vocabulary (for en_US)
    voc = load_json(os.path.join(dir_name, "langs", "en_US.json"))
    if not voc:
        eprint("Failed loading vocabulary")
        sys.exit(1)

    shell_data = load_shell_data()
    lang = os.getenv("LANG").split(".")[0] if not shell_data["interface-locale"] else shell_data["interface-locale"]
    # translate if translation available
    if lang != "en_US":
        loc_file = os.path.join(dir_name, "langs", "{}.json".format(lang))
        if os.path.isfile(loc_file):
            # localized vocabulary
            loc = load_json(loc_file)
            if not loc:
                eprint("Failed loading translation into '{}'".format(lang))
            else:
                for key in loc:
                    voc[key] = loc[key]


def signal_handler(sig, frame):
    global sig_dwl
    desc = {2: "SIGINT", 15: "SIGTERM", 10: "SIGUSR1"}
    if sig == 2 or sig == 15:
        print("Terminated with {}".format(desc[sig]))
        if tray_available:
            sni_system_tray.deinit_tray()
        Gtk.main_quit()
    elif sig == sig_dwl:
        refresh_dwl()
    else:
        return


def rt_sig_handler(sig, frame):
    print("{} RT signal received".format(sig))
    for executor in common.executors_list:
        if executor.use_sigrt and executor.sigrt == sig:
            eprint("Refreshing {} on signal {}".format(executor.name, sig))
            executor.refresh()

    for win in panel_windows_hide_show_sigs:
        if sig == panel_windows_hide_show_sigs[win]:
            if win.is_visible():
                win.hide()
            else:
                win.show()


def restart():
    subprocess.Popen(restart_cmd, shell=True)


def hypr_watcher():
    import socket

    # /tmp/hypr moved to $XDG_RUNTIME_DIR/hypr in #5788
    xdg_runtime_dir = os.getenv("XDG_RUNTIME_DIR")
    hypr_dir = f"{xdg_runtime_dir}/hypr" if xdg_runtime_dir and os.path.isdir(
        f"{xdg_runtime_dir}/hypr") else "/tmp/hypr"

    client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    client.connect(f"{hypr_dir}/{his}/.socket2.sock")

    while True:
        datagram = client.recv(2048)
        e_full_string = datagram.decode('utf-8', errors='replace').strip()
        lines = e_full_string.splitlines()

        event_names = []
        for line in lines:
            event_names.append(line.split(">>")[0])
        # print(f"events: {event_names}")

        for event_name in event_names:
            if common_settings["restart-on-display"] and (event_name in ["monitoradded", "monitorremoved"]):
                print("Received event '{}'; restart in {} ms.".format(event_name, common_settings["restart-delay"]))
                GLib.timeout_add(common_settings["restart-delay"], restart, priority=GLib.PRIORITY_HIGH)

            if event_name in ["activespecial",
                              "activewindow",
                              "activewindowv2",
                              "changefloatingmode",
                              "closewindow",
                              "createworkspace",
                              "destroyworkspace",
                              "focusedmon",
                              "monitoradded",
                              "movewindow",
                              "openwindow",
                              "windowtitle",
                              "workspace"]:

                # print(f">>> refreshing on {event_name}")
                monitors, workspaces, clients, activewindow, activeworkspace = h_modules_get_all()
                for item in common.h_taskbars_list:
                    GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow)

                for item in common.h_workspaces_list:
                    GLib.timeout_add(0, item.refresh, monitors, workspaces, clients, activewindow, activeworkspace)
                break

            elif event_name == "submap":
                for item in common.h_submaps_list:
                    GLib.timeout_add(0, item.refresh)


def niri_watcher():
    import socket

    client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    client.connect(niri_sock)
    client.sendall(b"\"EventStream\"\n")

    buffer = ""
    while True:
        chunk = client.recv(1024).decode('utf-8', errors='replace')
        if not chunk:
            break
        buffer += chunk
        while "\n" in buffer:
            line, buffer = buffer.split("\n", 1)
            line = line.strip()
            if not line:
                continue
            try:
                message = json.loads(line)
                # event_name = next(iter(message))
                # event_data = message[event_name]
                # print(f"[{event_name}]")

                outputs, workspaces, windows, focused_window = niri_get_all()
                # print(f"{outputs}\n{workspaces}\n{windows}\n{focused_window}")
                for item in common.niri_taskbars_list:
                    GLib.timeout_add(0, item.refresh, outputs, workspaces, windows, focused_window)

            except json.JSONDecodeError as e:
                print("Failed to decode JSON:", e)

def on_i3ipc_event(i3conn, event):
    if common_settings["restart-on-display"]:
        num = num_active_outputs(i3conn.get_outputs())
        if num > common.outputs_num:
            print("Number of outputs increased ({}); restart in {} ms.".format(
                num, common_settings["restart-delay"]))
            GLib.timeout_add(common_settings["restart-delay"],
                             restart,
                             priority=GLib.PRIORITY_HIGH)
        common.outputs_num = num

    GLib.idle_add(hide_controls_popup, priority=GLib.PRIORITY_HIGH)


def hide_controls_popup():
    for item in common.controls_list:
        if item.popup_window.get_visible():
            item.popup_window.hide_and_clear_tag()


def refresh_dwl(*args):
    if len(common.dwl_instances) > 0:
        dwl_data = load_json(common.dwl_data_file)
        if dwl_data:
            for item in common.dwl_instances:
                item.refresh(dwl_data)


def instantiate_content(panel, container, content_list, icons_path=""):
    check_key(panel, "position", "top")
    check_key(panel, "items-padding", 0)

    for item in content_list:
        # list initial data for Hyprland modules
        if his:
            if "hyprland-workspaces" in content_list or "hyprland-taskbar" in content_list:
                monitors, workspaces, clients, activewindow, activeworkspace = h_modules_get_all()
            else:
                monitors, workspaces, clients, activewindow, activeworkspace = {}, {}, {}, {}, {}

        if item == "sway-taskbar":
            if "sway-taskbar" in panel:
                if sway:
                    check_key(panel["sway-taskbar"], "all-outputs", False)
                    if panel["sway-taskbar"]["all-outputs"] or "output" not in panel:
                        taskbar = SwayTaskbar(panel["sway-taskbar"], common.i3, panel["position"],
                                              icons_path=icons_path)
                    else:
                        taskbar = SwayTaskbar(panel["sway-taskbar"], common.i3, panel["position"],
                                              display_name="{}".format(panel["output"]), icons_path=icons_path)

                    container.pack_start(taskbar, False, False, panel["items-padding"])
                else:
                    eprint("'sway-taskbar' ignored")
            else:
                eprint("'sway-taskbar' not defined in this panel instance")

        if item == "sway-workspaces":
            if sway:
                if "sway-workspaces" in panel:
                    workspaces = SwayWorkspaces(panel["sway-workspaces"], common.i3, icons_path=icons_path)
                    container.pack_start(workspaces, False, False, panel["items-padding"])
                else:
                    print("'sway-workspaces' not defined in this panel instance")
            else:
                eprint("'sway-workspaces' ignored")

        if item == "scratchpad":
            if sway:
                # Added in v0.1.3, so may be undefined in user's config.
                if item not in panel:
                    panel["scratchpad"] = {}
                scratchpad = Scratchpad(common.i3, common.i3.get_tree(), panel[item], panel["output"],
                                        icons_path=icons_path)
                container.pack_start(scratchpad, False, False, panel["items-padding"])
            else:
                eprint("'scratchpad' ignored")

        if item == "sway-mode":
            if sway:
                if item not in panel:
                    panel["sway-mode"] = {}
                sway_mode = SwayMode(common.i3, panel[item], icons_path=icons_path)
                container.pack_start(sway_mode, False, False, panel["items-padding"])
            else:
                eprint("'sway-mode' ignored")

        if item == "hyprland-taskbar":
            if "hyprland-taskbar" in panel:
                if his:
                    check_key(panel["hyprland-taskbar"], "all-outputs", False)
                    if panel["hyprland-taskbar"]["all-outputs"] or "output" not in panel:
                        taskbar = HyprlandTaskbar(panel["hyprland-taskbar"], panel["position"], monitors, workspaces,
                                                  clients, activewindow, icons_path=icons_path)
                    else:
                        taskbar = HyprlandTaskbar(panel["hyprland-taskbar"], panel["position"], monitors, workspaces,
                                                  clients, activewindow, display_name="{}".format(panel["output"]),
                                                  icons_path=icons_path)

                    common.h_taskbars_list.append(taskbar)
                    container.pack_start(taskbar, False, False, panel["items-padding"])
                else:
                    eprint("'hyprland-taskbar' ignored (HIS unknown).")

        if item == "niri-taskbar":
            if niri_sock:
                outputs, workspaces, windows, focused_window = {}, {}, {}, {}
                if "niri-taskbar" in panel:
                    if niri_sock:
                        check_key(panel, "niri-taskbar", {})
                        check_key(panel["niri-taskbar"], "all-outputs", False)
                        if panel["niri-taskbar"]["all-outputs"] or "output" not in panel:
                            taskbar = NiriTaskbar(panel["niri-taskbar"], panel["position"], outputs, workspaces,
                                                      windows, focused_window, icons_path=icons_path)
                        else:
                            taskbar = NiriTaskbar(panel["niri-taskbar"], panel["position"], outputs, workspaces,
                                                      windows, focused_window, display_name="{}".format(panel["output"]),
                                                      icons_path=icons_path)

                        common.niri_taskbars_list.append(taskbar)
                        container.pack_start(taskbar, False, False, panel["items-padding"])
                    else:
                        eprint("'niri-taskbar' ignored (NIRI_SOCKET unknown).")

        if item == "hyprland-workspaces":
            if his:
                if "hyprland-workspaces" in panel:
                    workspaces = HyprlandWorkspaces(panel["hyprland-workspaces"], panel["output"], monitors, workspaces, clients,
                                                    activewindow, activeworkspace, icons_path=icons_path)
                    container.pack_start(workspaces, False, False, panel["items-padding"])
                    common.h_workspaces_list.append(workspaces)
                else:
                    print("'hyprland-workspaces' not defined in this panel instance")
            else:
                eprint("'hyprland-workspaces' ignored")

        if item == "hyprland-submap":
            if his:
                if item in panel:
                    h_submap = HyprlandSubmap(panel[item], icons_path=icons_path)
                else:
                    h_submap = HyprlandSubmap({}, icons_path=icons_path)
                container.pack_start(h_submap, False, False, panel["items-padding"])
                common.h_submaps_list.append(h_submap)
            else:
                eprint("HyprlandSubmap module only works on Hyprland, ignoring")

        if item == "keyboard-layout":
            if his or sway or niri_sock:
                if "keyboard-layout" not in panel:
                    panel["keyboard-layout"] = {}
                kb_layout = KeyboardLayout(panel["keyboard-layout"], icons_path)
                container.pack_start(kb_layout, False, False, panel["items-padding"])
            else:
                eprint("KeyboardLayout module does not yet support your compositor")

        if item == "pinned":
            if "pinned" not in panel:
                panel["pinned"] = {}
            pinned = Pinned(panel["pinned"], icons_path)
            container.pack_start(pinned, False, False, panel["items-padding"])

        if "button-" in item:
            if item in panel:
                button = CustomButton(panel[item], icons_path)
                container.pack_start(button, False, False, panel["items-padding"])
            else:
                print("'{}' not defined in this panel instance".format(item))

        if "executor-" in item:
            if item in panel:
                executor = Executor(panel[item], icons_path, item)
                container.pack_start(executor, False, False, panel["items-padding"])
                common.executors_list.append(executor)
            else:
                print("'{}' not defined in this panel instance".format(item))

        if item == "clock":
            if item in panel:
                clock = Clock(panel[item], icons_path=icons_path)
                container.pack_start(clock, False, False, panel["items-padding"])
            else:
                clock = Clock({})
                container.pack_start(clock, False, False, 0)

        if item == "playerctl":
            if item in panel:
                playerctl = Playerctl(panel[item], voc, icons_path)
                container.pack_start(playerctl, False, False, panel["items-padding"])
            else:
                print("'{}' not defined in this panel instance".format(item))

        if item == "openweather":
            if "python-requests" in common.commands and common.commands["python-requests"]:
                if item in panel:
                    openweather = OpenWeather(panel[item], voc, icons_path=icons_path)
                    container.pack_start(openweather, False, False, panel["items-padding"])
            else:
                eprint("OpenWeather module needs the 'python-requests' package")

        if item == "brightness-slider":
            if item in panel:
                brightness_slider = BrightnessSlider(panel[item], icons_path)
                container.pack_start(brightness_slider, False, False, panel["items-padding"])

        if item == "cpu-avg":
            cpu_avg = CpuAvg()
            container.pack_start(cpu_avg, False, False, panel["items-padding"])

        if item == "random-wallpaper":
            if item not in panel:
                panel[item] = {}
            cpu_avg = RandomWallpaper(panel[item], voc, icons_path)
            container.pack_start(cpu_avg, False, False, panel["items-padding"])

        if item == "dwl-tags":
            if os.path.isfile(common.dwl_data_file):
                if "dwl-tags" not in panel:
                    panel["dwl-tags"] = {}

                dwl_tags = DwlTags(panel["output"], panel["dwl-tags"])
                common.dwl_instances.append(dwl_tags)
                container.pack_start(dwl_tags, False, False, panel["items-padding"])
                dwl_data = load_json(common.dwl_data_file)
                if dwl_data:
                    dwl_tags.refresh(dwl_data)
            else:
                eprint("{} data file not found".format(common.dwl_data_file))

        if item == "tray" and tray_available:
            tray_settings = {}
            if "tray" in panel:
                tray_settings = panel["tray"]
            tray = sni_system_tray.Tray(tray_settings, panel["position"], icons_path)
            common.tray_list.append(tray)
            container.pack_start(tray, False, False, panel["items-padding"])


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c",
                        "--config",
                        type=str,
                        default="config",
                        help="config filename (in {}/)".format(common.config_dir))

    parser.add_argument("-s",
                        "--style",
                        type=str,
                        default="style.css",
                        help="css filename (in {}/)".format(common.config_dir))

    parser.add_argument("-sigdwl",
                        type=int,
                        default=10,
                        help="signal to refresh dwl-tags module; default: 10 (SIGUSR1)")

    parser.add_argument("-r",
                        "--restore",
                        action="store_true",
                        help="restore default config files")

    parser.add_argument("-v",
                        "--version",
                        action="version",
                        version="%(prog)s version {}".format(__version__),
                        help="display version information")

    args = parser.parse_args()

    # Kill running instances, if any
    own_pid = os.getpid()
    running_instances = []

    for proc in psutil.process_iter(['pid', 'name']):
        name = proc.info['name']
        if name and "nwg-panel" in name and "-con" not in name:
            pid = proc.info['pid']
            if pid != own_pid:
                running_instances.append(pid)
                print(f"Running instance found, PID {pid}, sending SIGINT")
                try:
                    os.kill(pid, signal.SIGINT)
                except ProcessLookupError:
                    continue  # Process already dead

    # Wait for clean shutdown (max 3 seconds)
    max_wait = 3.0
    interval = 0.2
    elapsed = 0.0

    while elapsed < max_wait:
        alive = [p for p in running_instances if psutil.pid_exists(p)]
        if not alive:
            break
        time.sleep(interval)
        elapsed += interval

    # SIGKILL if still alive
    for pid in running_instances:
        if psutil.pid_exists(pid):
            print(f"PID {pid} still alive after {elapsed:.1f}s, sending SIGKILL")
            try:
                os.kill(pid, signal.SIGKILL)
            except ProcessLookupError:
                pass

    # Fallback: kill by PID file (should never happen)
    pid_file = os.path.join(temp_dir(), "nwg-panel.pid")
    if os.path.isfile(pid_file):
        try:
            pid = int(load_text_file(pid_file))
            if psutil.pid_exists(pid):
                print(f"Unnamed instance found via PID file, killing PID {pid}")
                os.kill(pid, signal.SIGKILL)
        except:
            pass

    # Warn if KDE's background daemon is running
    for proc in psutil.process_iter(['name']):
        if proc.info['name'] in ('kded5', 'kded6'):
            eprint(f"Warning: '{proc.info['name']}' is running and may block the system tray.")
            eprint("> See https://github.com/Alexays/Waybar/issues/3468 for details.")
            break

    save_string(str(own_pid), pid_file)

    common.config_dir = get_config_dir()

    load_vocabulary()

    global common_settings
    cs_file = os.path.join(common.config_dir, "common-settings.json")
    if not os.path.isfile(cs_file):
        common_settings = {
            "restart-on-display": True,
            "restart-delay": 500,
            "processes-backgroud-only": True,
            "processes-own-only": True
        }
        save_json(common_settings, cs_file)
    else:
        common_settings = load_json(cs_file)

    print("Common settings", common_settings)

    cache_dir = get_cache_dir()
    if cache_dir:
        common.dwl_data_file = os.path.join(cache_dir, "nwg-dwl-data")
        scratchpad_file = os.path.join(temp_dir(), "nwg-scratchpad")
        if os.path.isfile(scratchpad_file):
            common.scratchpad_cons = load_json(scratchpad_file)
            eprint("Loaded scratchpad info", common.scratchpad_cons)
    else:
        eprint("Couldn't determine cache directory")

    global sig_dwl
    sig_dwl = args.sigdwl

    catchable_sigs = set(signal.Signals) - {signal.SIGKILL, signal.SIGSTOP, signal.SIGCHLD}
    for sig in catchable_sigs:
        try:
            signal.signal(sig, signal_handler)
        except Exception as exc:
            eprint("{} subscription error: {}".format(sig, exc))

    for sig in range(signal.SIGRTMIN, signal.SIGRTMAX + 1):
        try:
            signal.signal(sig, rt_sig_handler)
        except Exception as exc:
            eprint("{} subscription error: {}".format(sig, exc))

    check_commands()
    print("Dependencies check:", common.commands)

    global restart_cmd
    restart_cmd = "nwg-panel -c {} -s {}".format(args.config, args.style)

    save_string("-c {} -s {}".format(args.config, args.style), os.path.join(local_dir(), "args"))

    config_file = os.path.join(common.config_dir, args.config)

    copy_files(os.path.join(dir_name, "icons_light"), os.path.join(common.config_dir, "icons_light"))
    copy_files(os.path.join(dir_name, "icons_dark"), os.path.join(common.config_dir, "icons_dark"))
    copy_files(os.path.join(dir_name, "icons_color"), os.path.join(common.config_dir, "icons_color"))
    # copy_files(os.path.join(dir_name, "langs"), os.path.join(common.config_dir, "langs"))
    copy_executors(os.path.join(dir_name, "executors"), os.path.join(common.config_dir, "executors"))
    copy_files(os.path.join(dir_name, "config"), common.config_dir, args.restore)
    copy_files(os.path.join(dir_name, "local"), local_dir())

    # tree = common.i3.get_tree() if sway else None
    common.outputs, common.mon_desc2output_name = list_outputs(sway=sway)
    print("Outputs:")
    for key in common.outputs:
        print(key, common.outputs[key])
    print(f"Descriptions: {common.mon_desc2output_name}")

    panels = load_json(config_file)

    screen = Gdk.Screen.get_default()
    provider = Gtk.CssProvider()
    style_context = Gtk.StyleContext()
    style_context.add_provider_for_screen(screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
    try:
        provider.load_from_path(os.path.join(common.config_dir, args.style))
    except Exception as e:
        eprint(e)

    # Controls background window (invisible): add style missing from the css file
    css = provider.to_string().encode('utf-8')
    css += b""" window#bcg-window { background-color: rgba(0, 0, 0, 0.2); } """
    provider.load_from_data(css)

    # Mirror bars to all outputs #48 (if panel["output"] == "All")
    to_remove = []
    to_append = []
    for panel in panels:
        check_key(panel, "output", "")
        check_key(panel, "monitor", "")
        check_key(panel, "run-if-output-exist", [])
        check_key(panel, "run-if-output-absent", [])
        if panel["monitor"]:
            try:
                panel["output"] = common.mon_desc2output_name[panel["monitor"]]
            except KeyError as err:
                eprint(f"Monitor description unknown: {err}")

        clones = []
        if panel["output"] == "All" and len(common.outputs) >= 1:
            to_remove.append(panel)
            for key in common.outputs.keys():
                clone = panel.copy()
                clone["output"] = key
                clones.append(clone)

            to_append = to_append + clones
        else:
            for output in panel["run-if-output-exist"]:
                if output not in common.outputs:
                    to_remove.append(panel)
            for output in panel["run-if-output-absent"]:
                if output in common.outputs:
                    to_remove.append(panel)

    for item in to_remove:
        if item in panels:
            panels.remove(item)

    panels = panels + to_append

    for panel in panels:
        monitor = None
        try:
            monitor = common.outputs[panel["output"]]["monitor"]
        except KeyError:
            pass

        if panel["output"] and not monitor:
            eprint("Couldn't assign a Gdk.Monitor to output '{}'".format(panel["output"]))
            continue

        check_key(panel, "icons", "")
        icons_path = ""
        if panel["icons"] == "light":
            icons_path = os.path.join(common.config_dir, "icons_light")
        elif panel["icons"] == "dark":
            icons_path = os.path.join(common.config_dir, "icons_dark")

        # This is to allow width "auto" value. Actually all non-numeric values will be removed.
        if "width" in panel and not isinstance(panel["width"], int):
            panel.pop("width")
            panel["width-as-percentage"] = False

        if panel["output"] in common.outputs or not panel["output"]:
            check_key(panel, "spacing", 6)
            check_key(panel, "css-name", "")
            check_key(panel, "padding-horizontal", 0)
            check_key(panel, "padding-vertical", 0)
            check_key(panel, "sigrt", 0)  # SIGRTMIN > hide_show_sig_num <= SIGRTMAX, (0 = disabled)
            check_key(panel, "use-sigrt", False)
            check_key(panel, "start-hidden", False)

            check_key(panel, "width-as-percentage", False)

            window = Gtk.Window()
            global panel_windows_hide_show_sigs
            if panel["use-sigrt"]:
                panel_windows_hide_show_sigs[window] = panel["sigrt"]
            else:
                panel_windows_hide_show_sigs[window] = 0

            if panel["css-name"]:
                window.set_property("name", panel["css-name"])

            if "output" not in panel or not panel["output"]:
                display = Gdk.Display.get_default()
                monitor = display.get_monitor(0)
                for key in common.outputs:
                    if common.outputs[key]["monitor"] == monitor:
                        panel["output"] = key

            # Width undefined or "auto"
            if "output" in panel and panel["output"] and "width" not in panel:
                panel["width"] = common.outputs[panel["output"]]["width"]

            # Width defined as percentage
            if "width" in panel and panel["width"] > 0 and panel["width-as-percentage"]:
                if panel["width"] > 100:
                    panel["width"] = 100
                panel["width"] = int(common.outputs[panel["output"]]["width"] * panel["width"] / 100)

            check_key(panel, "width", 0)
            w = panel["width"]

            check_key(panel["controls-settings"], "window-width", 0)
            controls_width = panel["controls-settings"]["window-width"] if panel["controls-settings"][
                                                                               "window-width"] > 0 else int(w / 5)
            check_key(panel, "height", 0)
            h = panel["height"]

            check_key(panel, "controls", "off")
            if panel["controls"]:
                check_key(panel, "controls-settings", {})

            if "controls-settings" in panel:
                controls_settings = panel["controls-settings"]
                check_key(controls_settings, "show-values", False)
                check_key(controls_settings, "window-margin", 0)

            check_key(panel, "menu-start", "off")
            if panel["menu-start"]:
                check_key(panel, "menu-start-settings", {})
                defaults = {
                    "cmd-lock": "swaylock -f -c 000000",
                    "cmd-logout": "swaymsg exit",
                    "cmd-restart": "systemctl reboot",
                    "cmd-shutdown": "systemctl -i poweroff",
                    "autohide": True,
                    "file-manager": "thunar",
                    "height": 0,
                    "icon-size-large": 32,
                    "icon-size-small": 16,
                    "icon-size-button": 16,
                    "margin-bottom": 0,
                    "margin-left": 0,
                    "margin-right": 0,
                    "margin-top": 0,
                    "padding": 2,
                    "terminal": "foot",
                    "width": 0
                }
                for key in defaults:
                    check_key(panel["menu-start-settings"], key, defaults[key])

            if panel["menu-start"] != "off":
                panel["menu-start-settings"]["horizontal-align"] = panel["menu-start"]

            Gtk.Widget.set_size_request(window, w, h)

            o = Gtk.Orientation.HORIZONTAL if panel["position"] == "top" or panel[
                "position"] == "bottom" else Gtk.Orientation.VERTICAL

            vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
            hbox = Gtk.Box(orientation=o, spacing=0)
            vbox.pack_start(hbox, True, True, panel["padding-vertical"])

            check_key(panel, "modules-left", [])
            check_key(panel, "modules-center", [])
            check_key(panel, "modules-right", [])
            check_key(panel, "homogeneous", False)

            inner_box = Gtk.Box(orientation=o, spacing=0)

            hbox.pack_start(inner_box, True, True, 0)
            hbox.set_property("margin-start", panel["padding-horizontal"])
            hbox.set_property("margin-end", panel["padding-horizontal"])
            hbox.set_property("margin-top", panel["padding-vertical"])
            hbox.set_property("margin-bottom", panel["padding-vertical"])

            left_box = Gtk.Box(orientation=o, spacing=panel["spacing"])
            left_box.set_property("name", "left-box")
            inner_box.pack_start(left_box, False, True, 0)
            if panel["controls"] and panel["controls"] == "left":
                monitor = None
                try:
                    monitor = common.outputs[panel["output"]]["monitor"]
                except KeyError:
                    pass

                cc = Controls(panel["controls-settings"], panel["position"], panel["controls"],
                              controls_width, monitor=monitor, icons_path=icons_path)
                common.controls_list.append(cc)
                left_box.pack_start(cc, False, False, 0)

                if common.commands["swaync"]:
                    if "swaync" not in panel:
                        panel["swaync"] = {}
                    sway_nc = SwayNC(panel["swaync"], icons_path, panel["position"])
                    left_box.pack_start(sway_nc, False, False, 0)

            if panel["menu-start"] == "left":
                ms = MenuStart(panel, icons_path=icons_path)
                left_box.pack_start(ms, False, False, 0)

            instantiate_content(panel, left_box, panel["modules-left"], icons_path=icons_path)
            print("left box created")

            center_box = Gtk.Box(orientation=o, spacing=panel["spacing"])
            center_box.set_property("name", "center-box")

            if panel["modules-center"] and panel["homogeneous"]:
                inner_box.set_center_widget(center_box)
            else:
                inner_box.pack_start(center_box, True, False, 0)

            check_key(panel, "modules-center", [])
            instantiate_content(panel, center_box, panel["modules-center"], icons_path=icons_path)
            print("center box created")

            right_box = Gtk.Box(orientation=o, spacing=panel["spacing"])
            right_box.set_property("name", "right-box")
            # Damn on the guy who invented `pack_start(child, expand, fill, padding)`!
            helper_box = Gtk.Box(orientation=o, spacing=0)
            helper_box.pack_end(right_box, False, False, 0)
            inner_box.pack_end(helper_box, False, True, 0)
            check_key(panel, "modules-right", [])
            instantiate_content(panel, right_box, panel["modules-right"], icons_path=icons_path)
            print("right box created")

            if panel["menu-start"] == "right":
                ms = MenuStart(panel["menu-start-settings"], icons_path=icons_path)
                right_box.pack_end(ms, False, False, 0)

            if panel["controls"] and panel["controls"] == "right":
                monitor = None
                try:
                    monitor = common.outputs[panel["output"]]["monitor"]
                except KeyError:
                    pass

                cc = Controls(panel["controls-settings"], panel["position"], panel["controls"],
                              controls_width, monitor=monitor, icons_path=icons_path)
                common.controls_list.append(cc)
                right_box.pack_end(cc, False, False, 0)

                if common.commands["swaync"]:
                    if "swaync" not in panel:
                        panel["swaync"] = {}

                    sway_nc = SwayNC(panel["swaync"], icons_path, panel["position"])
                    right_box.pack_end(sway_nc, False, False, 0)

            window.add(vbox)

            GtkLayerShell.init_for_window(window)
            GtkLayerShell.set_namespace(window, "nwg-panel")

            monitor = None
            try:
                monitor = common.outputs[panel["output"]]["monitor"]
            except KeyError:
                pass

            check_key(panel, "layer", "top")
            o = panel["output"] if "output" in panel else "undefined"
            m = panel["monitor"] if "monitor" in panel else "undefined"
            print("Panel '{}': output: {}, monitor: {}, position: {}, layer: {}, width: {}, height: {}".format(
                panel["name"], o, m,
                panel["position"],
                panel["layer"],
                panel["width"],
                panel["height"]))

            if monitor:
                GtkLayerShell.set_monitor(window, monitor)

            check_key(panel, "exclusive-zone", True)
            if panel["exclusive-zone"]:
                GtkLayerShell.auto_exclusive_zone_enable(window)

            layers = {"background": GtkLayerShell.Layer.BACKGROUND,
                      "bottom": GtkLayerShell.Layer.BOTTOM,
                      "top": GtkLayerShell.Layer.TOP,
                      "overlay": GtkLayerShell.Layer.OVERLAY}

            GtkLayerShell.set_layer(window, layers[panel["layer"]])

            """if panel["layer"] == "top":
                GtkLayerShell.set_layer(window, GtkLayerShell.Layer.TOP)
            else:
                GtkLayerShell.set_layer(window, GtkLayerShell.Layer.BOTTOM)"""

            check_key(panel, "margin-top", 0)
            GtkLayerShell.set_margin(window, GtkLayerShell.Edge.TOP, panel["margin-top"])

            check_key(panel, "margin-bottom", 0)
            GtkLayerShell.set_margin(window, GtkLayerShell.Edge.BOTTOM, panel["margin-bottom"])

            if panel["position"] == "top":
                GtkLayerShell.set_anchor(window, GtkLayerShell.Edge.TOP, 1)
            elif panel["position"] == "bottom":
                GtkLayerShell.set_anchor(window, GtkLayerShell.Edge.BOTTOM, 1)
            elif panel["position"] == "left":
                GtkLayerShell.set_anchor(window, GtkLayerShell.Edge.LEFT, 1)
            elif panel["position"] == "right":
                GtkLayerShell.set_anchor(window, GtkLayerShell.Edge.RIGHT, 1)

            if panel["use-sigrt"] and panel["start-hidden"]:
                window.hide()
            else:
                window.show_all()

    if sway:
        common.outputs_num = num_active_outputs(common.i3.get_outputs())
    else:
        common.outputs, common.mon_desc2output_name = list_outputs(sway=sway, silent=True)
        common.outputs_num = len(common.outputs)

    if sway:
        # Notice: Don't use Event.OUTPUT, it's not supported on old sway releases.
        common.i3.on(Event.WORKSPACE, on_i3ipc_event)
        common.i3.on(Event.WINDOW, on_i3ipc_event)

        # We monitor i3ipc events in a separate thread, and callbacks will also
        # be executed there. Hence, UI operations MUST be scheduled by
        # Gdk.threads_add_*() or their GLib counterpart to make them happen in
        # Gtk's main loop.
        thread = threading.Thread(target=common.i3.main, daemon=True)
        thread.start()

    if his:
        if len(common.h_taskbars_list) > 0 or len(common.h_workspaces_list) > 0:
            print("his: '{}', starting hypr_watcher".format(his))
            # read from Hyprland socket2 on another thread
            thread = threading.Thread(target=hypr_watcher, daemon=True)
            thread.daemon = True
            thread.start()

    if niri_sock:
        thread = threading.Thread(target=niri_watcher, daemon=True)
        thread.daemon = True
        thread.start()

    if tray_available and len(common.tray_list) > 0:
        sni_system_tray.init_tray(common.tray_list)

    Gtk.main()


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3

import os
import subprocess

import gi

from nwg_panel.tools import check_key, update_image, load_text_file, cmd_through_compositor, get_cache_dir, eprint

gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')

from gi.repository import Gtk, Gio


def get_app_dirs():
    desktop_dirs = []

    home = os.getenv("HOME")
    xdg_data_home = os.getenv("XDG_DATA_HOME")
    xdg_data_dirs = os.getenv("XDG_DATA_DIRS") if os.getenv("XDG_DATA_DIRS") else "/usr/local/share/:/usr/share/"

    if xdg_data_home:
        desktop_dirs.append(os.path.join(xdg_data_home, "applications"))
    else:
        if home:
            desktop_dirs.append(os.path.join(home, ".local/share/applications"))

    for d in xdg_data_dirs.split(":"):
        desktop_dirs.append(os.path.join(d, "applications"))

    # Add flatpak dirs if not found in XDG_DATA_DIRS
    flatpak_dirs = [os.path.join(home, ".local/share/flatpak/exports/share/applications"),
                    "/var/lib/flatpak/exports/share/applications"]
    for d in flatpak_dirs:
        if d not in desktop_dirs:
            desktop_dirs.append(d)

    return desktop_dirs


def parse_desktop_file(file_path, lang):
    try:
        with open(file_path, "r", encoding="utf-8") as f:
            lines = f.read().splitlines()
    except UnicodeDecodeError:
        eprint(f"Warning: Invalid .desktop file '{file_path}'")
        return None
    except OSError as e:
        eprint(f"Warning: Unable to read .desktop file '{file_path}': {e}")
        return None

    icon, _exec, name = "", "", ""
    for line in lines:
        if line.startswith("[") and line != "[Desktop Entry]":
            break
        if line.upper().startswith("ICON"):
            icon = line.split("=", 1)[1].strip()
        elif line.upper().startswith("EXEC"):
            _exec = line.split("=", 1)[1].strip()
        elif line.upper().startswith("NAME") and not name:
            name = line.split("=", 1)[1].strip()
        elif line.upper().startswith(f"NAME[{lang.upper()}]"):
            name = line.split("=", 1)[1].strip()

    if "%" in _exec:
        _exec = _exec.split("%")[0].strip()

    return icon, _exec, name


def launch(widget, cmd):
    cmd = cmd_through_compositor(cmd)
    print(f"Executing: {cmd}")
    subprocess.Popen('{}'.format(cmd), shell=True)


class Pinned(Gtk.EventBox):
    def __init__(self, settings, icons_path):
        Gtk.EventBox.__init__(self)
        self.file_monitor = None
        self.icons_path = icons_path

        check_key(settings, "limit", 0)
        check_key(settings, "icon-size", 16)
        check_key(settings, "angle", 0.0)
        self.settings = settings

        self.set_property("name", "root-pinned")

        lang = os.environ.get('LANG', 'en_US.UTF-8')  # fallback to default if not set
        self.lang = lang.split('.')[0].split('_')[0]

        self.desktop_dirs = get_app_dirs()

        self.desktop_ids = []
        self.cache_file_path = os.path.join(get_cache_dir(), "nwg-pin-cache")
        if os.path.exists(self.cache_file_path):
            cache_content = load_text_file(self.cache_file_path)
            if cache_content:
                self.desktop_ids = cache_content.splitlines()
                if self.desktop_ids:
                    self.box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
                    self.add(self.box)
                    if settings["angle"] != 0.0:
                        self.box.set_orientation(Gtk.Orientation.VERTICAL)

                    self.build_box()
            else:
                eprint(f"Nothing found in cache: '{self.cache_file_path}'")
        else:
            eprint(f"{self.cache_file_path} file not found")

        self.setup_file_monitor()

    def build_box(self):
        for widget in self.box.get_children():
            widget.destroy()

        counter = 0
        for desktop_id in self.desktop_ids:
            for desktop_dir in self.desktop_dirs:
                desktop_file = os.path.join(desktop_dir, desktop_id)

                if os.path.exists(desktop_file):
                    icon_name, exec, name = parse_desktop_file(desktop_file, lang=self.lang)
                    if icon_name and exec:
                        image = Gtk.Image()
                        update_image(image, icon_name, self.settings["icon-size"], self.icons_path)

                        btn = Gtk.Button()
                        btn.set_image(image)
                        btn.set_property("name", "pinned-button")
                        btn.set_tooltip_text(name)

                        btn.connect("clicked", launch, exec)
                        self.box.pack_start(btn, False, False, 0)
                    counter += 1
                    break

                if 0 < self.settings["limit"] <= counter:
                    break

        self.show_all()

    def setup_file_monitor(self):
        file = Gio.File.new_for_path(self.cache_file_path)
        self.file_monitor = file.monitor_file(Gio.FileMonitorFlags.NONE, None)
        self.file_monitor.connect("changed", self.on_file_changed)

    def on_file_changed(self, monitor, file, other_file, event_type):
        if event_type in [Gio.FileMonitorEvent.CHANGES_DONE_HINT, Gio.FileMonitorEvent.CREATED]:
            print(f"{self.cache_file_path} changed, rebuilding box...")
            cache_content = load_text_file(self.cache_file_path)
            if cache_content:
                self.desktop_ids = cache_content.splitlines()
            else:
                self.desktop_ids = []
            self.build_box()



#!/usr/bin/python3

import os
import time
import subprocess


# You may either use the full icon(s) path here, like e.g.:
# "/home/piotr/.config/nwg-panel/icons_light/arch-linux.svg"
# or just give the icon name, like below.

# The icon name must either exist in your icon theme, or you may place `icon_name.svg`
# custom files in '~/.config/nwg-panel/icons_light/' and '~/.config/nwg-panel/icons_dark/'.

# This script needs the `au.sh` helper on path or in the same directory. See comments in `au.sh`.

def main():
    # Avoid checking on each panel restart: check if 15 minutes passed.
    # Adjust the time (in seconds) to your liking.
    # Make sure if the path below matches your temp directory.
    file = "/tmp/arch-updates"
    
    if os.path.isfile(file):
        if int(time.time() - os.stat(file).st_mtime) > 900:
            arch, aur = check_updates()
            save_string("{},{}".format(arch, aur), file)
        else:
            try:
                vals = load_string(file).split(",")
                arch, aur = int(vals[0]), int(vals[1])
            except:
                arch, aur = 0, 0
    else:
        arch, aur = check_updates()
        save_string("{},{}".format(arch, aur), file)

    if arch > 0 and aur > 0:
        print("software-update-urgent")
        print("{}/{}".format(arch, aur))
    elif arch > 0:
        print("software-update-available")
        print("{}".format(arch))
    elif aur > 0:
        print("software-update-available")
        print("{}".format(aur))


def save_string(string, file):
    try:
        file = open(file, "wt")
        file.write(string)
        file.close()
    except:
        print("Error writing file '{}'".format(file))


def load_string(path):
    try:
        with open(path, 'r') as file:
            data = file.read()
            return data
    except:
        return ""


def check_updates():
    arch, aur = 0, 0
    try:
        arch = len(subprocess.check_output(["checkupdates"]).decode("utf-8").splitlines())
    except:
        pass
    try:
        aur = len(subprocess.check_output(["trizen", "-Qqu", "-a"]).decode("utf-8").splitlines())
    except:
        pass

    return arch, aur


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import os

from gi.repository import Gtk

import subprocess

from nwg_panel.tools import check_key, update_image, cmd_through_compositor


class MenuStart(Gtk.Button):
    def __init__(self, panel, icons_path=""):
        Gtk.Button.__init__(self)
        self.set_always_show_image(True)
        self.panel = panel
        check_key(panel, "menu-start-settings", {})
        self.settings = panel["menu-start-settings"]

        defaults = {
            "cmd-lock": "swaylock -f -c 000000",
            "cmd-logout": "swaymsg exit",
            "cmd-restart": "systemctl reboot",
            "cmd-shutdown": "systemctl -i poweroff",
            "autohide": False,
            "file-manager": "thunar",
            "menu-start": "right",
            "icon-size-button": 16,
            "icon-size-large": 32,
            "icon-size-small": 16,
            "margin-bottom": 0,
            "margin-left": 0,
            "margin-right": 0,
            "margin-top": 0,
            "output": "",
            "padding": 2,
            "terminal": "foot",
            "position": "bottom",
            "run-through-compositor": True,
            "hover-opens-submenu": False,
            "click-outside-closes": False
        }

        for key in defaults:
            if key not in self.settings:
                self.settings[key] = defaults[key]

        self.set_property("name", "button-start")

        image = Gtk.Image()
        update_image(image, "nwg-shell", self.settings["icon-size-button"], icons_path)
        self.set_image(image)

        self.connect("clicked", self.on_click)

        self.show()

    def on_click(self, button):
        cmd = "nwg-menu"

        print(self.settings)
        if self.settings["cmd-lock"] != "swaylock -f -c 000000":
            cmd += " -cmd-lock '{}'".format(self.settings["cmd-lock"])
        if self.settings["cmd-logout"] != "swaymsg exit":
            cmd += " -cmd-logout '{}'".format(self.settings["cmd-logout"])
        if self.settings["cmd-restart"] != "systemctl reboot":
            cmd += " -cmd-restart '{}'".format(self.settings["cmd-restart"])
        if self.settings["cmd-shutdown"] != "systemctl -i poweroff":
            cmd += " -cmd-shutdown '{}'".format(self.settings["cmd-shutdown"])
        if self.settings["autohide"]:
            cmd += " -d"
        if self.settings["file-manager"] != "thunar":
            cmd += " -fm {}".format(self.settings["file-manager"])
        if self.panel["menu-start"] == "right":
            cmd += " -ha {}".format(self.panel["menu-start"])
        if self.settings["icon-size-large"] != 32:
            cmd += " -isl {}".format(self.settings["icon-size-large"])
        if self.settings["icon-size-small"] != 16:
            cmd += " -iss {}".format(self.settings["icon-size-small"])
        if self.settings["margin-bottom"] > 0:
            cmd += " -mb {}".format(self.settings["margin-bottom"])
        if self.settings["margin-left"] > 0:
            cmd += " -ml {}".format(self.settings["margin-left"])
        if self.settings["margin-right"] > 0:
            cmd += " -mr {}".format(self.settings["margin-right"])
        if self.settings["margin-top"] > 0:
            cmd += " -mt {}".format(self.settings["margin-top"])
        if self.panel["output"]:
            cmd += " -o {}".format(self.panel["output"])
        if self.settings["padding"] != 2:
            cmd += " -padding {}".format(self.settings["padding"])
        if self.settings["terminal"] != "foot":
            cmd += " -term {}".format(self.settings["terminal"])
        if self.panel["position"] != "bottom":
            cmd += " -va {}".format(self.panel["position"])
        if self.settings["hover-opens-submenu"]:
            cmd += " -t"
        if self.settings["click-outside-closes"]:
            cmd += " -k"

        if self.settings["run-through-compositor"]:
            if os.getenv("SWAYSOCK"):
                cmd += " -wm sway"
            elif os.getenv("HYPRLAND_INSTANCE_SIGNATURE"):
                cmd += " -wm hyprland"

        cmd = cmd_through_compositor(cmd)

        print(f"Executing: {cmd}")
        subprocess.Popen('{}'.format(cmd), shell=True)
